# Copyright (C) 2023 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import cryptocode
import os
import getpass
from .logging import get_logger
from pykeepass import PyKeePass

logger = get_logger(__name__)


class KeepassCli:
    _name = "Keepass"

    def __init__(self, configurator, keepass_password=''):
        self.configurator = configurator
        self.init_keepass(keepass_password)

    def get_keepass_params(self):
        bitwarden_params = ['keepass_path', 'keepass_group']
        return {param: self.configurator.config.get(param) for param in bitwarden_params}

    def init_keepass(self, keepass_password):
        password = keepass_password or os.environ.get('KEEPASS_PASSWORD', '')
        self._keepass_password = password
        params = self.get_keepass_params()
        self.keepass_path = keepass_password or params.get('keepass_path', '')
        self.keepass_group = keepass_password or params.get('keepass_group', '')

    def get_keepass(self, filename):
        return PyKeePass(os.getenv("HOME") + "/" + filename, password=self._keepass_password)

    def get_keepass_entry(self, filename, group, entry):
        if not self._keepass_password:
            self._keepass_password = getpass.getpass('Keepass Password: ')
        kp = self.get_keepass(filename)
        kgroup = kp.find_groups(name=group, first=True)
        if kgroup:
            kentry = kp.find_entries(title=entry, group=kgroup, recursive=False, first=True)
            if kentry:
                return kentry
            else:
                logger.error("Cannot find %s in %s" % (entry, group))
                raise Exception("Cannot find %s in %s" % (entry, group))
        else:
            logger.error("Cannot find %s" % group)
            raise Exception("Cannot find %s" % group)

    def get_keepass_entry_value(self, filename, group, entry, field, default='admin'):
        try:
            if filename and group and entry:
                entry = self.get_keepass_entry(filename, group, entry)
            elif filename and group and not entry:
                entry = self.get_keepass_entry(self.keepass_path, filename, group)
            elif filename and not group and not entry:
                entry = self.get_keepass_entry(self.keepass_path, self.keepass_group, filename)
            else:
                raise ValueError('No keepass entry found')
        except Exception as err:
            logger.error(type(err).__name__)
            return False
        if entry:
            return eval('entry.' + field)
        return default

    def get_keepass_password(self, filename='', group='', entry=''):
        return self.get_keepass_entry_value(filename, group, entry, 'password')

    def get_keepass_password_crypt(self, filename='', group='', entry='', key=''):
        return cryptocode.encrypt(self.get_keepass_entry_value(filename, group, entry, 'password'), key)

    def get_keepass_user(self, filename='', group='', entry=''):
        return self.get_keepass_entry_value(filename, group, entry, 'username')

    def get_keepass_url(self, filename='', group='', entry=''):
        kp_url = self.get_keepass_entry_value(filename, group, entry, 'url')
        if kp_url and kp_url[-1] == '/':
            kp_url = kp_url[:-1]
        return kp_url



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict
from . import base


class OdooSystemParameter(base.OdooModule):
    _name = "System Parameter"

    def __init__(self, configurator, auto_apply=True):
        self.auto_apply = auto_apply
        super(OdooSystemParameter, self).__init__(configurator)

    def apply(self, datas=None):
        super(OdooSystemParameter, self).apply()
        for key in datas or self._datas:
            if isinstance(self._datas.get(key), dict) or isinstance(self._datas.get(key), OrderedDict):
                system_parameters = self._datas.get(key).get('system_parameter', {})
                if system_parameters:
                    self.logger.info("\t- %s" % key)
                    self.execute(system_parameters)

    def execute_script_config(self, datas=False):
        datas = datas if datas else self._datas
        for key in datas:
            if isinstance(datas.get(key), dict) or isinstance(datas.get(key), OrderedDict):
                values = datas.get(key).get('system_parameter', {})
                if values:
                    self.logger.info("\tSystem Parameter - %s" % key)
                    self.execute(values)

    def execute(self, datas):
        for key in datas:
            if isinstance(datas.get(key), dict) or isinstance(datas.get(key), OrderedDict):
                self.logger.info("\t\t* %s" % key)
                self.odoo_system_parameter(datas.get(key))

    def odoo_system_parameter(self, system_parameter):
        for key in system_parameter:
            if isinstance(system_parameter[key], str) and system_parameter[key].startswith('get_'):
                system_parameter[key] = self.safe_eval(system_parameter[key])
        domain = [['key', '=', system_parameter['key']]]
        parameter_id = self.search_read('ir.config_parameter', domain, fields=['id', 'key', 'value'],
                                   context=self._context, limit=1)
        if parameter_id:
            parameter_id = parameter_id[0]
            self.execute_odoo('ir.config_parameter', 'write',
                              [parameter_id['id'],
                               {'value': system_parameter['value']}
                               ], {'context': self._context})
        else:
            parameter_id = self.execute_odoo('ir.config_parameter', 'create',
                                             [{'key': system_parameter['key'],
                                               'value': system_parameter['value']}],
                                             {'context': self._context})



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import connection
from . import modules
from . import translations
from . import datas
from . import system_parameter
from . import config
from . import roles
from . import defaults
from . import users
from . import account
from . import website
from . import imports
from . import import_configurator
from . import call



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict

from . import base


class OdooDefaults(base.OdooModule):
    _name = "Defaults"

    def apply(self):
        super(OdooDefaults, self).apply()

        for key in self._datas:
            if isinstance(self._datas.get(key), dict) or isinstance(self._datas.get(key), OrderedDict):
                defaults = self._datas.get(key).get('defaults', {})
                if defaults:
                    self.logger.info("\t- %s" % key)
                    self.pre_config(defaults)
                    self.odoo_defaults(defaults)

    def odoo_defaults(self, defaults):
        if defaults:
            for default in defaults:
                self.logger.info("\t\t* %s" % default)
                default_value = defaults[default]['value']
                if type(default_value) is str and default_value.startswith('get_'):
                    default_value = self.safe_eval(default_value)
                self.execute_odoo('ir.default', 'set',
                                  [defaults[default]['model'],
                                   defaults[default]['field'],
                                   default_value,
                                   False, False, defaults[default].get('condition', False)], {'context': self._context})



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2025 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from .configurator import Configurator
import argparse

import importlib.metadata
import pathlib
import toml

source_location = pathlib.Path(__file__).parent.parent
if (source_location.parent / "pyproject.toml").exists():
    with open(source_location.parent / "pyproject.toml", "rt") as f:
        __version__ = toml.load(f)['project']['version']
else:
    __version__ = importlib.metadata.version("odoo-configurator")


def main():
    parser = argparse.ArgumentParser(description='Odoo Configurator')
    parser.add_argument('paths', metavar='files', type=open, nargs='+', help='Config Files To load')
    parser.add_argument('--update', action='store_true', help='Update Mode')
    parser.add_argument('--install', action='store_true', help='Install Mode')
    parser.add_argument('--debug', action='store_true', help='Debug log')
    parser.add_argument('--debug_xmlrpc', action='store_true', help='Debug log xmlrpc')
    parser.add_argument('--keepass', type=str, help='Keepass password')
    parser.add_argument('--lang', type=str, help='Language code (e.g. fr_FR) to be used in context for all API calls')
    parser.add_argument('--slack-token', type=str, help='Slack token')
    args = parser.parse_args()
    c = Configurator(**dict(args._get_kwargs()), version=__version__)
    c.show()
    log = c.start()


if __name__ == '__main__':
    main()



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright 2024 Scalizer (<https://www.scalizer.fr>)
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import time
from collections import OrderedDict

from . import base


class OdooModules(base.OdooModule):
    _name = "Modules"
    _modules_cache = {}
    _modules_updated = False
    _modules_theme_cache = {}
    _uninstalled_modules_cache = {}

    def apply(self):
        pass  # For remove standard log

    def pre_update_config_modules(self):
        if self._pre_datas:
            self.install_modules(self._pre_datas)
            self.update_modules(self._pre_datas)
            self.uninstall_modules(self._pre_datas)

    def install_config_modules(self, datas=None):
        if not datas:
            datas = self._datas
        self.install_modules(datas)
        self.update_modules(datas)
        self.uninstall_modules(datas)

    def install_modules(self, config):
        if config.get('modules', []):
            self.install_odoo(config.get('modules', []))
        for key in config:
            if isinstance(config.get(key), dict) or isinstance(config.get(key), OrderedDict):
                modules = config.get(key).get('modules', [])
                if modules:
                    self.logger.info("\tInstall Modules - %s" % key)
                    self.install_odoo(modules)

    def update_modules(self, config):
        self.update_module_odoo(config.get('updates', []))
        for key in config:
            if isinstance(config.get(key), dict) or isinstance(config.get(key), OrderedDict):
                modules = config.get(key).get('updates', [])
                if modules:
                    self.logger.info("\tUpdate Modules - %s" % key)
                    self.update_module_odoo(modules)

    def uninstall_modules(self, config):
        if config.get('uninstall_modules', []):
            self.uninstall_odoo(config.get('uninstall_modules', []))
        for key in config:
            if isinstance(config.get(key), dict) or isinstance(config.get(key), OrderedDict):
                modules = config.get(key).get('uninstall_modules', [])
                if modules:
                    self.logger.info("\tUninstall Modules - %s" % key)
                    self.uninstall_odoo(modules)

    def update_list(self):
        if not self._modules_updated:
            self.execute_odoo('ir.module.module', 'update_list', [])
            self._modules_updated = True
        return {i['name']: {'state': i['state'], 'id': i['id']} for i in
                self.execute_odoo('ir.module.module', 'search_read',
                                  [[], ['name', 'state'], 0, 0, "id"],
                                  {'context': self._context})}

    def install_odoo(self, modules):
        if not self._modules_cache:
            self._modules_cache = self.update_list()

        if modules:
            to_install = []
            missing_modules = []
            for module in modules:
                self.logger.info('\t\t* %s' % module)
                if self._modules_cache.get(module):
                    if self._modules_cache.get(module).get('state') != 'installed':
                        to_install.append(self._modules_cache.get(module).get('id'))
                else:
                    missing_modules.append(module)

            if missing_modules:
                self.logger.error("\tModules not found : %s " % (", ".join(missing_modules)))

            if to_install:
                self.execute_odoo('ir.module.module', 'button_immediate_install', [to_install])
                self.logger.info('\t\t* Done')

    def update_module_odoo(self, modules):
        if modules:
            self.logger.info('\tUpdate %s', modules)
            self.execute_odoo('ir.module.module', 'button_immediate_upgrade',
                              [[self._connection.get_id_from_xml_id("base.module_" + m) for m in modules]])
            self.logger.info('\t\t* Done')

    def uninstall_odoo(self, modules):
        if not self._uninstalled_modules_cache:
            self._uninstalled_modules_cache = self.update_list()

        if modules:
            to_uninstall = []
            for module in modules:
                self.logger.info('\t\t* %s' % module)
                if self._uninstalled_modules_cache.get(module).get('state') != 'uninstalled':
                    to_uninstall.append(self._uninstalled_modules_cache.get(module).get('id'))

            for m in to_uninstall:
                self.execute_odoo('ir.module.module', 'button_immediate_uninstall', [m], no_raise=True)
                time.sleep(3)
                self.logger.info('\t\t* %s uninstalled' % m)

    def install_odoo_theme(self, module):
        if not self._modules_theme_cache:
            self.update_list()
            self._modules_theme_cache = {i['name']: {'state': i['state'], 'is_installed_on_current_website': i[
                'is_installed_on_current_website'], 'id': i['id']} for i in
                                         self.execute_odoo('ir.module.module', 'search_read',
                                                           [[], ['name', 'state', 'is_installed_on_current_website'], 0,
                                                            0, "id"], {'context': self._context})}

        if module:
            if not self._modules_theme_cache.get(module).get('is_installed_on_current_website'):
                self.execute_odoo('ir.module.module', 'button_choose_theme',
                                  [[self._modules_theme_cache.get(module).get('id')]])



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright 2024 Scalizer (<https://www.scalizer.fr>)
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from datetime import date, datetime
import glob
import logging
from packaging.version import Version
import os.path
import sys

import hiyapyco

from .apps import connection
from .apps import account
from .apps import config
from .apps import system_parameter
from .apps import datas
from .apps import defaults
from .apps import imports
from .apps import python_script
from .apps import modules
from .apps import roles
from .apps import translations
from .apps import users
from .apps import website
from .apps import mattermost
from .apps import slack
from .apps import call
from .apps import import_configurator

from .import_manager import ImportManager
from .logging import get_logger
from .odoo_connection import OdooConnection
from .keepass import KeepassCli
from .bitwarden import Bitwarden
from .utils import Utils

logger = get_logger(__name__)


def get_config_from_files(files):
    return hiyapyco.load(files, method=hiyapyco.METHOD_MERGE, interpolate=True,
                         failonmissingfiles=True, loglevel='INFO')


class Configurator:
    mode = ["config"]
    debug = False
    connection = None
    import_manager = None
    keepass_cli = None
    bitwarden_cli = None
    config = dict()
    pre_update_config = dict()
    xmlid_cache = dict()
    release_directory = ''
    clear_release_directory = False
    paths = list()
    slack_token = ''
    start_time = datetime.now()
    version = ''

    def __init__(self, paths=None, install=False, update=False, debug=False, debug_xmlrpc=False, **kwargs):
        """
        paths (list): List of file paths to load configuration from.
        install (bool): Whether to install mode is enabled.
        update (bool): Whether to update mode is enabled.
        debug (bool): Whether to enable debug logging.
        debug_xmlrpc (bool): Whether to enable debug logging for XML-RPC.
        **kwargs: Additional keyword arguments:
            - keepass (str): Keepass password.
            - slack_token (str): Slack token.
            - config_dict (dict): Dictionary containing configuration data.
        """
        if paths is None:
            paths = []
        self.configurator_dir = os.path.dirname(sys.argv[0])
        paths = [i.name if hasattr(i, 'name') else i for i in paths]
        if install:
            self.mode.append('install')
        if update:
            self.mode.append('update')

        self.version = kwargs.get('version', '')

        self.slack_token = os.environ.get('SLACK_TOKEN') or kwargs.get('slack_token', '')
        self.lang = kwargs.get('lang', 'fr_FR')
        self.debug = debug
        self.debug_xmlrpc = debug_xmlrpc
        if paths:
            self.paths = paths
            self.config, self.pre_update_config = self.parse_config()
        else:
            self.config = kwargs.get('config_dict', {})
        self.log_history = []
        logger.setLevel(logging.DEBUG if debug else logging.INFO)

        self.utils = Utils(self)
        self.keepass_cli = KeepassCli(self, keepass_password=kwargs.get('keepass', ''))
        self.bitwarden_cli = Bitwarden(self)
        self.prepare_odoo_connection()
        self.import_manager = ImportManager(self)

    def prepare_odoo_connection(self):
        odoo_params = self.get_odoo_auth_params()
        connection.OdooConnection(self).pre_config(odoo_params)
        self.connection = OdooConnection(
            odoo_params['url'],
            odoo_params['dbname'],
            odoo_params['username'],
            odoo_params['password'],
            version=self.config.get('version', False),
            http_user=odoo_params.get('http_user'),
            http_password=odoo_params.get('http_password'),
            createdb=odoo_params.get('create_db'),
            debug_xmlrpc=self.debug_xmlrpc,
            lang=self.lang,
            configurator=self,
        )

    def get_odoo_auth_params(self):
        if self.config.get('auth') and self.config.get('auth').get('odoo'):
            return self.config['auth']['odoo']
        else:
            return {
                'url': os.environ.get('ODOO_URL'),
                'dbname': os.environ.get('ODOO_DB'),
                'username': os.environ.get('ODOO_USER'),
                'password': os.environ.get('ODOO_PASSWORD'),
            }

    def parse_config(self):
        count_inherit = 0
        count_pre_update = 0
        config_files = []
        pre_update_config = None
        parsed_config = get_config_from_files(self.paths)

        while len(parsed_config.get("pre_update", [])) != count_pre_update:
            count_pre_update = len(parsed_config.get("pre_update", []))
            config_files = self.get_files_path(parsed_config['pre_update'])
            logger.info("Pre Update Loading %s" % (",".join(config_files)))
            pre_update_config = get_config_from_files(config_files)
            pre_update_config['auth'] = parsed_config['auth']

        while len(parsed_config.get("inherits", [])) != count_inherit:
            count_inherit = len(parsed_config.get("inherits", []))
            inherit_files = self.get_files_path(parsed_config['inherits'])
            config_files = self.paths + inherit_files
            parsed_config = get_config_from_files(config_files)
        logger.info("Configuration Loading %s" % (",".join(config_files)))

        if parsed_config.get('release_directory'):
            self.release_directory = parsed_config['release_directory']
            release_files = self.get_release_files()
            config_files += release_files
            logger.info("Release Configuration Loading %s" % (",".join(release_files)))
            parsed_config = get_config_from_files(config_files)
        if parsed_config.get('clear_release_directory'):
            self.clear_release_directory = parsed_config.get('clear_release_directory')

        if parsed_config.get('configurator_version'):
            version = parsed_config['configurator_version']
            version = str(version) if not isinstance(version, str) else version
            if Version(version) > Version(self.version):
                msg = "The yml configuration requires Odoo Configurator version >= %s (Current version==%s)"
                logger.error(msg % (version, self.version))
                exit(1)

        self.parse_scripts(parsed_config)

        return parsed_config, pre_update_config

    def parse_scripts(self, parsed_config):
        parsed_config['scripts'] = []
        count_script = 0
        while len(parsed_config.get("script_files", [])) != count_script:
            count_script = len(parsed_config.get("script_files", []))
            script_files = self.get_files_path(parsed_config['script_files'])

            for script_file in script_files:
                parsed_script = get_config_from_files([script_file])
                if parsed_script.get('title'):
                    parsed_script['title'] = '%s : %s' % (os.path.basename(script_file),
                                                          parsed_script.get('title'))
                else:
                    parsed_script['title'] = os.path.basename(script_file)
                if 'script_files' in parsed_script:
                    self.parse_scripts(parsed_script)
                parsed_config['scripts'].append(parsed_script)

    def get_release_files(self):
        files = []
        release_dir = self.release_directory
        if not os.path.isdir(release_dir):
            release_dir = os.path.join(os.path.dirname(sys.argv[1]), release_dir)
        if os.path.isdir(release_dir):
            files = glob.glob(os.path.join(release_dir, '*.yml'))
        return files

    def backup_release_directory(self):
        release_directory = self.utils.get_dir_full_path(self.release_directory)
        if self.clear_release_directory and os.path.isdir(release_directory):
            bak_dir = os.path.join(release_directory, 'bak')
            if not os.path.isdir(bak_dir):
                os.mkdir(bak_dir)
            release_bak_dir = os.path.join(bak_dir, str(date.today()))
            if not os.path.isdir(release_bak_dir):
                os.mkdir(release_bak_dir)
            for file in self.get_release_files():
                bak_file = os.path.join(release_bak_dir, os.path.basename(file))
                os.rename(file, bak_file)

    def get_files_path(self, files):
        res = []
        template_dirs = [
            os.path.join(self.configurator_dir, 'templates'),
            os.path.join(self.configurator_dir, 'src/templates'),
            os.path.join(os.path.dirname(__file__), '../templates'),
        ]
        for file in files:
            if os.path.isfile(file):
                res.append(file)
            else:
                file_found = ''
                for template_dir in template_dirs:
                    file_path = os.path.join(template_dir, file)
                    if os.path.isfile(file_path):
                        file_found = file_path
                        res.append(file_path)
                        continue
                if file_found:
                    continue

                for path in self.paths:
                    file_path = os.path.join(os.path.dirname(path), file)
                    if os.path.isfile(file_path):
                        file_found = file_path
                        continue

                if file_found:
                    res.append(file_found)
                else:
                    logger.info("File not found: %s" % file)
        return res

    def get_log(self):
        return "\n".join(self.log_history)

    def get_external_config_xmlid_cache(self):
        domain = [['module', '=', 'external_config']]
        xmlid_datas = self.connection.execute_odoo('ir.model.data', 'search_read', [domain, ['name', 'res_id']], {'context': {}})
        for xmlid_data in xmlid_datas:
            self.xmlid_cache['external_config.%s' % xmlid_data['name']] = xmlid_data['res_id']

    def show(self):
        pass
        # logger.info('show')
        # logger.info(pformat(self.config))

    def start(self):
        self.start_time = datetime.now()
        self.slack = slack.Slack(self)
        self.slack.init_slack_client()
        translations.OdooTranslations(self)
        modules_manager = modules.OdooModules(self)
        modules_manager.pre_update_config_modules()
        datas_manager = datas.OdooDatas(self)
        datas_manager.execute_pre_update_config_datas()
        modules_manager.install_config_modules()
        datas_manager.execute_update_config_datas()
        config.OdooConfig(self)
        system_parameter.OdooSystemParameter(self)
        roles.OdooRoles(self)
        defaults.OdooDefaults(self)
        users.OdooUsers(self)
        account.OdooAccount(self)
        website.OdooWebsite(self)
        imports.OdooImports(self)
        python_script.PythonScript(self)
        import_configurator.ImportConfigurator(self)
        mattermost.Mattermost(self)
        call.OdooCalls(self)
        self.backup_release_directory()

        duration = datetime.now() - self.start_time
        minutes = round(duration.total_seconds() / 60, 2) if duration else 0
        message = '%s updated in %s minutes' % (self.config.get('name'), minutes)
        if self.slack.slack_client:
            self.slack.send_message(message, message_type='valid', title='Odoo Configurator Done')
        logger.info(message)

        return self.get_log()



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base
from collections import OrderedDict


class OdooCalls(base.OdooModule):
    _name = "Calls"
    _key = "call"

    def apply(self):
        super(OdooCalls, self).apply()
        calls = []
        for key in self._datas:
            if isinstance(self._datas.get(key), dict) or isinstance(self._datas.get(key), OrderedDict):
                if self._datas.get(key).get(self._key, {}):
                    calls = {}
                    for call in self._datas.get(key).get(self._key, {}):
                        print(self.install_mode() , self._datas.get(key).get(self._key, {}).get(call).get(
                            'on_install_only', False))
                        if not self.install_mode() and self._datas.get(key).get(self._key, {}).get(call).get('on_install_only', False):
                            pass
                        else:
                            calls[call] = self._datas.get(key).get(self._key, {}).get(call)

        for call in calls:
            call_name = call
            call = calls[call]
            context = self._context
            context.update(call.get('context',{}))
            if call.get('model',{}) and call.get('method',{}):
                args = call.get('args',[])
                args_ok = []
                for i in args or []:
                    if isinstance(i,OrderedDict):
                        args_ok.append(dict(i))
                    else:
                        args_ok.append(i)
                self.logger.info('\t- Execute %s : env[\'%s\'].%s(%s,context=%s)' % (call_name, call.get('model', {}), call.get('method', {}), args_ok,call.get('context', {})))
                res = self.execute_odoo(call.get('model',{}), call.get('method',{}), args_ok, {'context': context})
                self.logger.info('\t- Call result %s' % (res))
            else:
                self.logger.info('\t- Call error %s' % (call_name))




# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from .logging import get_logger
import os
import sys

class Utils:
    """This class is intended to provide utility functions which are not related to a specific configurator app.
    Methods starting with 'get_' will automatically be callable from within yaml files,
    as per similar methods in OdooConnection Class (get_ref, get_record, etc.) """

    def __init__(self, configurator):
        self.configurator = configurator
        self.logger = get_logger("Utils".ljust(20))

    def get_env_var(self, var_name):
        var = os.environ.get(var_name)
        if not var:
            self.logger.warning(f"{var} Environment Variable not found or empty")
        return var

    @staticmethod
    def get_file_full_path(path):
        if not path:
            return ''
        param_path = path
        if not os.path.isfile(path):
            path = os.path.join(os.path.dirname(sys.argv[1]), param_path)
        if not os.path.isfile(path):
            path = os.path.join(os.path.dirname(sys.argv[1]), 'datas', param_path)
        if not os.path.isfile(path):
            raise FileNotFoundError('%s not found!' % param_path)
        return path

    @staticmethod
    def get_dir_full_path(path):
        if not path:
            return ''
        param_path = path
        if not os.path.isdir(path):
            path = os.path.join(os.path.dirname(sys.argv[1]), param_path)
        if not os.path.isdir(path):
            path = os.path.join(os.path.dirname(sys.argv[1]), 'datas', param_path)
        if not os.path.isdir(path):
            raise NotADirectoryError('%s not found!' % param_path)
        return path


# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import base64
import logging
import os.path
import re
import sys
import pickle
import ssl
import xmlrpc.client
from pprint import pformat
from .logging import get_logger
from s6r_odoo import OdooConnection as Orm
from .utils import Utils as utils

import requests
from bs4 import BeautifulSoup

CACHE = "/tmp/.configurator_cache"

METHODE_MAPPING = {
    15: [('get_object_reference', 'check_object_reference')]
}


class OdooConnection:
    _cache = {}

    def __init__(self, url, dbname, user, password, version=False, http_user=None, http_password=None, createdb=False,
                 debug_xmlrpc=False, configurator=None, **kwargs):
        self.logger = get_logger("Odoo Connection".ljust(15))
        if debug_xmlrpc:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self._url = url
        self._dbname = dbname
        self._user = user
        self._password = password
        self._http_user = http_user
        self._http_password = http_password
        self._version = version
        self._configurator = configurator
        self._context = {'lang': kwargs.get('lang') or 'fr_FR', 'noupdate': True}
        self.xmlid_cache = configurator.xmlid_cache if configurator else {}
        # noinspection PyProtectedMember,PyUnresolvedReferences
        self._insecure_context = ssl._create_unverified_context()
        self._load_cache()
        self._compute_url()
        try:
            self.odoo = Orm(self._url, self._dbname, self._user, self._password, debug_xmlrpc=debug_xmlrpc, lang=kwargs.get('lang','fr_FR'))
        except ConnectionError as err:
            self.logger.error(err)
            exit(1)
        except xmlrpc.client.Fault as err:
            self.logger.error(err.faultString)
            exit(1)
        except Exception as err:
            self.logger.error(err)
        if createdb:
            self._create_db()
        if self.odoo.uid:
            self.common = self.odoo.common
            self.object = self.odoo.object
            self.uid = self.odoo.uid
        else:
            self._prepare_connection()

    @property
    def context(self):
        return self._context

    def _load_cache(self):
        self._cache = {}
        if os.path.isfile(CACHE):
            with open(CACHE, "rb") as f:
                self._cache = pickle.loads(f.read())

    def _save_cache(self):
        with open(CACHE, "wb") as f:
            f.write(pickle.dumps(self._cache))

    def _compute_url(self):
        if self._http_user or self._http_password:
            self._url = self._url.replace('https://', 'https://%s:%s@' % (self._http_user, self._http_password))

    def _get_xmlrpc_method(self, method):
        new_method = method
        for v in METHODE_MAPPING:
            if self._version >= v:
                for i in METHODE_MAPPING[v]:
                    if i[0] == method:
                        new_method = i[1]
        return new_method

    def _create_db(self):
        post = {
            'master_pwd': "admin123",
            'name': self._dbname,
            'login': self._user,
            'password': self._password,
            'phone': '',
            'lang': 'fr_FR',
            'country_code': 'fr',
        }
        session = requests.Session()
        session.verify = False
        r = session.post(url=self._url + "/web/database/create", params=post)
        soup = BeautifulSoup(r.text, 'html.parser')
        alert = soup.find('div', attrs={"class": u"alert alert-danger"})
        if alert:
            self.logger.debug(self._url + "/web/database/create")
            self.logger.debug(post)
            self.logger.debug(alert.get_text())
            if "already exists" not in alert.text:
                raise Exception(alert.text)

    def _prepare_connection(self):
        self.logger.info("Prepare connection %s %s %s" % (self._url, self._dbname, self._user))
        self.common = xmlrpc.client.ServerProxy('{}/xmlrpc/2/common'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.object = xmlrpc.client.ServerProxy('{}/xmlrpc/2/object'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.uid = self.common.authenticate(self._dbname, self._user, self._password, {})
        self.logger.debug('Connection uid : %s' % self.uid)
        if not self.uid:
            raise Exception("Connection Error to %s %s %s" % (self._url, self._dbname, self._user))

    def execute_odoo(self, *args, no_raise=False, retry=0):
        self.logger.debug("*" * 50)
        self.logger.debug("Execute odoo :")
        self.logger.debug("\t Model : %s" % (args[0]))
        self.logger.debug("\t Method : %s" % (args[1]))
        self.logger.debug("\t " + "%s " * (len(args) - 2) % args[2:])
        self.logger.debug("*" * 50)
        try:
            res = self.object.execute_kw(self._dbname, self.uid, self._password, *args)
            return res
        except Exception as e:
            if no_raise:
                return
            if args[1] == 'search' and 'BaseModel.search() takes from 2 to 5 positional arguments' in str(e):
                self.logger.error(e.faultString)
                self.logger.error('Consider to add the version parameter in configuration')
                exit(1)

            if retry <= 3:
                return self.execute_odoo(*args, retry=retry+1)
            else:
                self.logger.error(pformat(args))
                if isinstance(e, xmlrpc.client.Fault):
                    self.logger.error(e.faultString)
                    exit(1)
                else:
                    self.logger.error(e, exc_info=True)
                raise e

    def search(self, model, domain, offset=0, limit=None, order=None, count=None, context=None):
        args = [domain, offset, limit, order]
        if self._version <= 16:
            args.append(count)
        return self.execute_odoo(model, 'search', args, {'context': context})

    def get_ref(self, external_id):
        if not self.xmlid_cache:  # load xml_id cache
            self._configurator.get_external_config_xmlid_cache()
        if external_id in self.xmlid_cache:
            return self.xmlid_cache[external_id]
        res = self.execute_odoo('ir.model.data',
                                self._get_xmlrpc_method('get_object_reference'),
                                external_id.split('.'))[1]
        self.logger.debug('Get ref %s > %s' % (external_id, res))
        self.xmlid_cache[external_id] = res
        return res

    def get_image_url(self, url):
        if 'image_url' not in self._cache:
            self._cache['image_url'] = {}
        if url not in self._cache['image_url']:
            self._cache['image_url'][url] = base64.b64encode(requests.get(url).content).decode("utf-8", "ignore")
            self._save_cache()
        return self._cache['image_url'][url]

    def get_image_local(self, path):
        path = utils.get_file_full_path(path)
        return base64.b64encode(open(path, "rb").read()).decode("utf-8", "ignore")

    @staticmethod
    def get_local_file(path, encode=False):
        path = utils.get_file_full_path(path)
        if encode:
            with open(path, "rb") as f:
                res = f.read()
            res = base64.b64encode(res).decode("utf-8", "ignore")
        else:
            with open(path, "r") as f:
                res = f.read()
        return res

    def get_country(self, code):
        return self.search('res.country', [('code', '=', code)],
                           limit=1, order='id', context=self._context)[0]

    def get_menu(self, website_id, url):
        return self.search('website.menu', [('website_id', '=', website_id), ('url', '=', url)],
                           limit=1, order='id', context=self._context)[0]

    def get_search_id(self, model, domain, order='asc'):
        res = self.search(model, domain, limit=1, order="id %s" % order, context=self._context)
        return res[0] if res else False

    def get_id_from_xml_id(self, xml_id, no_raise=False):
        if '.' not in xml_id:
            xml_id = "external_config." + xml_id
        try:
            object_reference = self._get_xmlrpc_method('get_object_reference')
            res_object_reference = self.execute_odoo('ir.model.data', object_reference, xml_id.split('.'),
                                                     {'context': {'active_test': False}}, no_raise=True)
            return res_object_reference[1] if res_object_reference else False
        except Exception as e:  # TODO : get true exception type and return False
            if no_raise:
                pass
            else:
                raise e

    def get_xml_id_from_id(self, model, res_id):
        try:
            domain = [('model', '=', model), ('res_id', '=', res_id)]
            res = self.execute_odoo('ir.model.data', 'search_read', [domain, ['module', 'name'], 0, 0, "id"],
                                    {'context': self._context})
            if not res:
                return
            ir_model_data = res[0]
            return "%s.%s" % (ir_model_data['module'], ir_model_data['name'])
        except Exception as e:  # TODO : get true exception type and return False
            raise e
            # return False

    def create_xml_id(self, module, model, name, res_id):
        name = re.sub('[^a-zA-Z0-9]', '_', name).lower()
        self.odoo.create('ir.model.data', {
            'module': module,
            'name': name,
            'model': model,
            'res_id': res_id
        })

    def set_active(self, is_active, model, domain, search_value_xml_id):
        if search_value_xml_id:
            object_id = self.get_id_from_xml_id(search_value_xml_id)
            domain = [(domain[0][0], domain[0][1], object_id)]
        object_ids = self.search(model,  domain, order='id', context=self._context)
        self.execute_odoo(model, 'write', [object_ids, {'active': is_active}], {'context': self._context})

    def read_search(self, model, domain, context=False):
        res = self.execute_odoo(model, 'search_read',
                                [domain],
                                {'context': context or self._context})
        return res

    def search_read(self, model, domain=[], fields=[], order=[], offset=0, limit=0, context=False):
        params = [domain, fields, offset, limit, order]
        res = self.execute_odoo(model, 'search_read', params, {'context': context or self._context})
        return res

    def get_record(self, model, rec_id, context=False):
        params = [[('id', '=', rec_id)]]
        res = self.execute_odoo(model, 'search_read',
                                params,
                                {'context': context or self._context})
        if res:
            return res[0]

    def default_get(self, model, field):
        res = self.execute_odoo(model, 'default_get', [field])
        return res



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import base64
from collections import OrderedDict
import os
from . import base

EXCLUDE_FIELDS = ['id', 'create_uid', 'create_date', 'write_uid', 'write_date', 'parent_path']


def compare_field(field):
    if field == 'name':
        res = 0
    elif field.endswith('_id'):
        res = 10
    elif field.endswith('_ids'):
        res = 20
    else:
        res = 99
    return res


def get_fle_type_from_binary(binary):
    if binary.startswith(b'\x89PNG\r\n\x1a\n\x00'):
        return 'png'
    if binary.startswith(b'\xff\xd8\xff'):
        return 'jpg'
    if binary.startswith(b'\x00\x00\x01\x00'):
        return 'ico'
    return 'bin'


def compare_xmlid(xmlid):
    module = xmlid.split('.')[0]
    if module == 'base':
        res = 'A'
    elif module.startswith('teclib') or module.startswith('hodei'):
        res = 'T'
    elif module.startswith('__'):
        res = 'Z'
    else:
        res = 'F'
    return res


def sort_fields(field_names):
    field_names = sorted(field_names, key=compare_field)
    return field_names


class ImportConfigurator(base.OdooModule):
    _name = "Import Configurator"

    rec_to_xmlid_cache = {}
    model_fields = {}
    display_name_prefix_fields = []

    def get_configurator_records(self, model, domain=[], excluded_fields=[], force_export_fields=[], order_by='',
                                 display_name_prefix_fields='', group_by=[], with_load=False, context={}, **kwargs):
        files = []
        self.display_name_prefix_fields = display_name_prefix_fields
        if kwargs.get('ids'):
            res_ids = [self.get_id_from_xml_id(xml_id) for xml_id in kwargs.get('ids')]
            self.logger.info("Import %s ids : %s", model, res_ids)
            domain = [('id', 'in', res_ids)]
        if not model:
            return '', []
        if 'create_xmlid' in kwargs:
            create_xmlid = kwargs.get('create_xmlid')
        else:
            create_xmlid = True
        if 'record_ids' in kwargs:
            records = kwargs.get('record_ids')
        else:
            records = self.odoo.model(model).search(domain, order=order_by, context=context)
        if not records:
            self.logger.info("No records to import for %s" % model)
            return '', []

        res = ''
        prev_record_group = ''
        model_id = self.odoo.model('ir.model').search([('model', '=', model)],
                                                      fields=['name'],
                                                      context=context)[0]
        self.load_model_fields(model)
        for i, record in enumerate(records):
            self.logger.info("Export %s : %s/%s", model, i+1, len(records))
            record_group = model_id['name'].title().replace(':', '')
            if group_by:
                if isinstance(record[group_by], list):
                    group_by_name = record[group_by][1]
                elif hasattr(record[group_by], 'name'):
                    group_by_name = record[group_by].name
                else:
                    group_by_name = record[group_by]
                record_group = '%s - %s' % (group_by_name, record_group)
            if prev_record_group != record_group:
                res += '\n%s:' % record_group
                if with_load:
                    res += '\n%s%s: %s' % (" " * 4, 'load', with_load)
                    res += '\n%s%s: %s' % (" " * 4, 'model', model)
                res += '\n%s%s:' % (" " * 4, 'datas')
                prev_record_group = record_group

            values = ""
            rec_name = self.get_record_name(record._values, model)
            prefix = self.get_record_prefix(record)
            if prefix:
                rec_name = '%s %s' % (prefix, rec_name)
            rec_name = rec_name.replace(':', ' ').replace('#', ' ').strip()

            xmlid = self.get_xmlid(model, record.id)
            if not xmlid:
                xmlid = self.compute_xml_id(record, model, create_xmlid=create_xmlid)
            res += '\n%s%s:' % (" " * 4 * 2, rec_name)
            res += '\n%s%s: %s' % (" " * 4 * 3, 'model', model)
            res += '\n%s%s: %s' % (" " * 4 * 3, 'force_id', xmlid)
            res += '\n%s%s:' % (" " * 4 * 3, 'values')

            field_names = sort_fields(kwargs.get('fields', [])) or sort_fields(self.model_fields.keys())
            for key in field_names:
                field = self.model_fields[key]['field']

                if not self.field_is_exclude(field, excluded_fields):
                    try:
                        default = self.model_fields[key]['default']
                        if (default != record[key] and record[key]) or key in force_export_fields:
                            field_value, files = self.get_field_value(record, field, files)
                            values += field_value
                    except Exception as e:
                        self.logger.error(e, exc_info=True)

            res += values
            res += '\n'

        return res, files

    def load_model_fields(self, model):
        self.model_fields = dict()
        fields = self.search_read('ir.model.fields', [('model', '=', model)])
        for field in fields:
            self.model_fields[field['name']] = {'field': field,
                                                'default': self.default_get(model, field['name'])}

    def field_is_exclude(self, field, excluded_fields=[]):
        field_name = field.get('name')
        if field_name in EXCLUDE_FIELDS:
            return True
        if field_name in excluded_fields:
            return True
        if field_name.startswith('__'):
            return True
        if not field.get('store', True):
            return True
        if field.get('related', False):
            return True

    def get_field_value(self, record, field, files):
        field_name = field.get('name')
        model = field.get('model')
        field_type = field.get('ttype')
        name = ''
        if field_type in ['one2many']:
            name = ''
        elif field_type in ['binary']:
            if field_name in ['image_1024', 'image_128', 'image_256', 'image_512']:
                name = ''
            binary_data = base64.b64decode(record[field_name])
            file_name = '%s_%s' % (model.replace('.', '_'), record.id)
            if model == 'ir.attachment' and '.' in record.name:
                ext = record.name.split('.')[-1]
            else:
                ext = get_fle_type_from_binary(binary_data)
            file_name = '%s.%s' % (file_name, ext)
            files.append((file_name, binary_data))
            name = "\n%s%s: %s" % (" " * 4 * 4, field_name, 'get_image_local("%s")' % file_name)
        elif field_type == 'many2many':
            rec_values = ''
            xmlid_list = [self.get_xmlid(field['relation'], val) for val in record[field_name].ids]
            for xmlid in xmlid_list:
                if xmlid:
                    rec_values += '\n%s- %s' % (" " * (2 + 4 * 4), xmlid)
            if rec_values:
                name = '\n%s%s/id:%s' % (" " * 4 * 4, field_name, rec_values)

        elif field_type in ['many2one']:
            if record[field_name]:
                xmlid = self.get_xmlid(field['relation'], record[field_name]['id'])
                if xmlid:
                    name = "\n%s%s/id: %s" % (" " * 4 * 4, field_name, xmlid)
            else:
                name = "\n%s%s.id: %s" % (" " * 4 * 4, field_name, 'False')
        elif field_type in ['char', 'date', 'datetime', 'selection']:
            val_str = record[field_name] and str(record[field_name]) or ''
            if '{{' in val_str:
                val_str = val_str.replace('{{', '\\{\\{')
                multiline_text = (" " * 4 * 5) + val_str.replace('\n', '\n%s' % (" " * 4 * 5))
                name = '\n%s%s: |\n%s' % (" " * 4 * 4, field_name, multiline_text)
            else:
                name = '\n%s%s: "%s"' % (" " * 4 * 4, field_name, val_str.replace('"', '\\"'))

        elif field_type in ['html', 'text']:
            multiline_text = (" " * 4 * 5) + record[field_name].replace('\n', '\n%s' % (" " * 4 * 5))
            name = '\n%s%s: |\n%s' % (" " * 4 * 4, field_name, multiline_text)
        else:
            name = "\n%s%s: %s" % (" " * 4 * 4, field_name, record[field_name])

        return name, files

    def get_record_name(self, record, model):
        try:
            if 'name' in record:
                if record.get('name'):
                    return record.get('name').title().replace(':', '')
                else:
                    return ''
            elif 'display_name' in record:
                return record.get('display_name').replace(':', '')
            elif 'key' in record:
                return record.get('key')
            elif 'ref' in record:
                return record.get('ref')
            else:
                return model + '_' + record.get('id')
        except Exception as e:
            self.logger.error(e, exc_info=True)
            pass
        return ''

    def get_record_prefix(self, record):
        prefix_words = []
        for field_name in self.display_name_prefix_fields:
            prefix = ''
            field_id = self.model_fields[field_name]['field']
            if field_id['ttype'] == 'many2one' and record[field_name]:
                related_record = self.search_read(field_id['relation'], [('id', '=', record[field_name].id)])
                prefix = related_record and related_record[0]['name']
            elif field_id['ttype'] == 'many2many' and record[field_name]:
                related_records = self.search_read(field_id['relation'], [('id', 'in', record[field_name].ids)])
                if related_records:
                    prefix = ' '.join([related_record['name'] for related_record in related_records])
            else:
                prefix = record[field_name]
            if prefix:
                prefix_words.append(prefix)
        return ' '.join(prefix_words)

    def get_xmlid(self, model, res_id):
        if (model, res_id) in self.rec_to_xmlid_cache:
            return self.rec_to_xmlid_cache[(model, res_id)]
        xmlid = self.get_xml_id_from_id(model, res_id)
        self.rec_to_xmlid_cache[(model, res_id)] = xmlid
        return xmlid

    def compute_xml_id(self, record, model, retry=0, create_xmlid=False):
        name = '%s_%s' % (model.replace('.', '_'), str(record['id']+retry).zfill(5))
        xml_id = 'external_config.%s' % name
        if not self.get_id_from_xml_id(xml_id, no_raise=True):
            if create_xmlid:
                self.create_xml_id('external_config', model, name, record['id'])
            return xml_id
        else:
            retry += 1
            return self.compute_xml_id(record, model, retry)

    def apply(self):
        super(ImportConfigurator, self).apply()
        for key in self._datas:
            if isinstance(self._datas.get(key), dict) or isinstance(self._datas.get(key), OrderedDict):
                action_name = 'import_configurator_model_file'
                if action_name in self._datas.get(key):
                    configurator_model_files = self._datas.get(key).get(action_name, {})
                    if configurator_model_files:
                        self.logger.info("\t- %s : %s" % (key, action_name))
                        self.import_configurator_model_file(configurator_model_files)

                action_name = 'import_configurator_module'
                if action_name in self._datas.get(key):
                    module = self._datas.get(key).get(action_name).get('module', '')
                    if module:
                        self.logger.info("\t- %s : %s (%s)" % (key, action_name, module))
                        self.import_configurator_module(module)

    def import_configurator_module(self, module):
        dest_path = os.path.join(os.path.dirname(self._configurator.paths[0]), 'config', module)
        if not os.path.isdir(dest_path):
            os.mkdir(dest_path)
        data_ids = self.odoo.search_ir_model_data([('module', '=', module)])
        if not data_ids:
            return
        models = data_ids.mapped('model')
        for model in models:
            model_data_ids = data_ids.filtered(model=model)
            record_ids = self.odoo.model(model).read(model_data_ids.mapped('res_id'))
            file_name = model.replace('.', '_')
            params = {'model': model,
                      'record_ids': record_ids}
            res, files = self.get_configurator_records(**params)
            open('%s/%s.yml' % (dest_path, file_name), 'w').write(res)

    def import_configurator_model_file(self, configurator_model_files):
        for configurator_model_file in configurator_model_files:
            model_file = configurator_model_files[configurator_model_file]
            model = model_file.get('model')
            domain = model_file.get('domain')
            ids = model_file.get('ids')
            order_by = model_file.get('order_by')
            group_by = model_file.get('group_by')
            context = dict(model_file.get('context', {}))
            load = model_file.get('load')
            force_export_fields = model_file.get('force_export_fields', [])
            excluded_fields = model_file.get('excluded_fields', [])
            create_xmlid = model_file.get('create_xmlid', True)
            fields = model_file.get('fields', [])
            display_name_prefix_fields = model_file.get('display_name_prefix_fields', [])
            file_name = model.replace('.', '_')
            dest_path = model_file.get('dest_path', '')
            dest_path = '%s/%s' % (os.path.dirname(self._configurator.paths[0]), dest_path or 'config')
            params = {'model': model,
                      'domain': domain,
                      'ids': ids,
                      'order_by': order_by,
                      'group_by': group_by,
                      'with_load': load,
                      'force_export_fields': force_export_fields,
                      'fields': fields,
                      'excluded_fields': excluded_fields,
                      'display_name_prefix_fields': display_name_prefix_fields,
                      'create_xmlid': create_xmlid,
                      'context': context}
            res, files = self.get_configurator_records(**params)
            open('%s/%s.yml' % (dest_path, file_name), 'w').write(res)

                # ##############################
                # #  CONFIGURATOR BINARY FILE  #
                # ##############################
                # action_name = 'import_configurator_binary_file'
                # configurator_model_files = self._datas.get(key).get(action_name, {})
                # if configurator_model_files:
                #     res = self.execute_odoo('ir.config_parameter', 'search_read',
                #                             [[('key', '=', 'odoo_configurator.access.token')],
                #                              ['value']],
                #                             {'context': self._context})
                #     if res:
                #         token = res[0].get('value', False)
                #     else:
                #         return
                #     url = '%s/teclib_configurator_generator/get_binary_file/%s' % \
                #           (self._configurator.connection._url, token)
                #     self.logger.info("\t- %s : %s" % (key, action_name))
                #     res = self.execute_odoo('ir.config_parameter', 'search_read',
                #                             [[('key', '=', 'odoo_configurator.access.token')],
                #                              ['value']],
                #                             {'context': self._context})
                #     if res:
                #         token = res[0].get('value', False)
                #         s = requests.Session()
                #         s.get(url='%s/web?db=%s' % (self._configurator.connection._url,
                #                                     self._configurator.connection._dbname))
                #
                #         for configurator_model_file in configurator_model_files:
                #             model_file = literal_eval(configurator_model_file)
                #             model = model_file[0]
                #             field = model_file[1]
                #             res_ids = model_file[2]
                #             for res_id in res_ids:
                #                 dest_path = os.path.dirname(self._configurator.paths[0]) + '/datas'
                #
                #                 r = s.get(url=url, params={'model': model, 'field': field, 'res_id': res_id})
                #
                #                 if r.headers['content-disposition']:
                #                     file_name = r.headers['content-disposition'].split('=')[1].split("''")[1]
                #                 else:
                #                     file_name = '%s_%s.bin' % (model.replace('.', '_'), res_id)
                #                 open('%s/%s' % (dest_path, file_name), 'wb').write(r.content)



# Copyright (C) 2025 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from . import base
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

MESSAGE_TYPE_EMOJIS = {
    'red': 'red_circle',  # 🔴
    'green': 'large_green_circle',  # 🟢
    'yellow': 'large_yellow_circle',  # 🟡
    'orange': 'large_orange_circle',  # 🟠
    'blue': 'large_blue_circle',  # 🔵
    'error': 'x',  # ❌
    'warning': 'warning',  # ⚠️
    'info': 'information_source',  # i️
    'valid': 'white_check_mark',  # ✅
    'start': 'arrow_forward',  # ▶️
}

class Slack(base.OdooModule):
    _name = "Slack"
    _key = "slack"

    slack_token = ''
    slack_channel = ''
    slack_client = None

    def apply(self):
        super(Slack, self).apply()

    def init_slack_client(self):
        if self._datas.get('no_notification', False):
            return
        self.slack_channel = self._datas.get('slack_channel', False)
        slack_token = self._configurator.slack_token or self._datas.get('slack_token', False)
        if slack_token and slack_token.startswith('get_'):
            self.slack_token = self.safe_eval(slack_token)
        else:
            self.slack_token = slack_token
        if self.slack_token:
            self.slack_client = WebClient(token=self.slack_token)

        if self.slack_client:
            self.logger.info('Slack client initialized')
            message = 'Starting Odoo Configurator : %s' % self.config.get('name')
            self.send_message(message, slack_channel=self.slack_channel, message_type='start')

    def send_message(self, message="This is a test! :tada:", slack_channel='', message_type='', title='', emoji=''):
        """
        Send a message to the specified Slack channel.
        :param message: The content of the message to be sent.
        :param slack_channel: The Slack channel where the message will be sent.
        :param message_type: The type of message. Available types: 'error', 'warning', 'info', 'valid', 'red', 'green',
        'yellow', 'orange', 'blue'.
        :param title: The title of the message.
        :param emoji: The Slack emoji code without the enclosing ":" e.g., "large_blue_circle".
        """
        message = self._prepare_message(message, message_type, title, emoji)
        if not self.slack_client:
            return
        try:
            self.logger.debug('Sending Slack message: %s' % message)
            channel = slack_channel or self.slack_channel
            response = self.slack_client.chat_postMessage(channel=channel, text=message)
        except SlackApiError as e:
            self.logger.error('Slack API error: %s' % e.response["error"])

    def _prepare_message(self, message, message_type, title, emoji):
        if title:
            message = f'*{title}*\n{message}'
        if message_type:
            emoji = MESSAGE_TYPE_EMOJIS.get(message_type, '')
        if emoji:
            message = f':{emoji}: {message}'

        return message



# Copyright (C) 2023 - Teclib'ERP (<https://www.teclib-erp.com>).
# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from collections import OrderedDict
from ast import literal_eval
import json

from . import base
from .config import OdooConfig
from .modules import OdooModules
from .users import OdooUsers
from .imports import OdooImports
from .system_parameter import OdooSystemParameter


def prepare_load_values(load_fields, fields, values):
    load_values = []
    for load_field in load_fields:
        field_name = load_field.split('.')[0].split('/')[0]
        if load_field in fields:
            value = values[fields.index(load_field)]
        elif field_name in fields:
            value = values[fields.index(field_name)]
        else:
            value = False
        load_values.append(value)

    return load_values


class OdooDatas(base.OdooModule):
    _name = "Datas"

    def apply(self):
        pass  # For remove standard log

    def execute(self, datas):
        for key in datas:
            values = datas.get(key)
            if isinstance(values, dict) or isinstance(values, OrderedDict):
                data = values.get('datas', {})
                if data:
                    self.logger.info("\tDatas - %s" % key)
                    self.odoo_datas(data, load=datas[key].get('load', False))

        scripts = datas.get('scripts', [])
        odoo_config = OdooConfig(self._configurator, auto_apply=False)
        odoo_system_parameter = OdooSystemParameter(self._configurator, auto_apply=False)
        odoo_modules = OdooModules(self._configurator)
        odoo_users = OdooUsers(self._configurator)
        odoo_imports = OdooImports(self._configurator, auto_apply=False)
        for script in scripts:
            self.logger.info("Script - %s" % script.get('title'))
            odoo_modules.install_config_modules(script)
            odoo_config.execute_script_config(script)
            odoo_system_parameter.execute_script_config(script)
            odoo_users.execute(script)
            odoo_imports.apply(script)
            self.execute(script)

    def execute_pre_update_config_datas(self):
        self.logger.info("Apply Pre-%s" % self._name)
        if self._pre_datas:
            self.execute(self._pre_datas)

    def execute_update_config_datas(self):
        self.logger.info("Apply %s" % self._name)
        self.execute(self._datas)

    def execute_config(self, config):
        if config:
            self.pre_config(config)
            self._connection.execute_config(config)

    def odoo_datas(self, datas, load=False):
        self.pre_config(datas)
        load = load or datas.pop('load', False)
        model = datas.pop('model', False)
        load_fields = []
        raw_load_values = []
        for data in datas:
            object_ids = False
            self.logger.info("\t\t* %s" % data)
            if not isinstance(datas[data], dict):
                return
            if 'install' not in self._mode and datas[data].get('on_install_only', False):
                return
            model = datas[data].get('model') or model
            field_key = datas[data].get('key', False)
            force_id = datas[data].get('force_id', False)
            delete_all = datas[data].get('delete_all', False)
            delete_domain = datas[data].get('delete_domain', False)
            delete_id = datas[data].get('delete_id', False)
            deactivate = datas[data].get('deactivate', False)
            activate = datas[data].get('activate', False)
            update_domain = datas[data].get('update_domain', False)
            search_value_xml_id = datas[data].get('search_value_xml_id', False)
            action_server_id = datas[data].get('action_server_id', False)
            function = datas[data].get('function', False)
            no_raise = datas[data].get('no_raise', False)
            context = dict(datas[data].get('context', {}))
            values = datas[data].get('values', {})
            languages = datas[data].get('languages', [self._context['lang']])

            # escape {{
            for val_key in values:
                val = values[val_key]
                if isinstance(val, str) and '\\{\\{' in val:
                    values[val_key] = val.replace('\\{\\{', '{{')

            config_context = self._context.copy()
            config_context.update(context)

            if delete_all or delete_domain:
                if 'install' not in self._mode and datas[data].get('on_install_only', False):
                    continue
                if delete_domain:
                    domain = self.eval_param_value(delete_domain)
                else:
                    domain = []
                object_ids = self.search(model, domain, order='id', context=config_context)
                for object_id in object_ids:
                    self.execute_odoo(model, 'unlink', [object_id])
                continue

            if delete_id:
                try:
                    object_id = self.get_id_from_xml_id(delete_id, no_raise=True)
                    if object_id:
                        self.execute_odoo(model, 'unlink', [object_id])
                except Exception as e:
                    self.logger.error(e)
                    pass
                continue

            if update_domain:
                domain = self.eval_param_value(update_domain)
                if search_value_xml_id:
                    object_id = self.get_id_from_xml_id(search_value_xml_id)
                    for condition in domain:
                        if condition[2] == 'search_value_xml_id':
                            condition[2] = object_id
                object_ids = self.search(model, domain, order='id', context=config_context)
                self.logger.debug("Update Domain %s %s" % (len(object_ids), model))
                self.execute_odoo(model, 'write', [object_ids, dict(values)], {'context': config_context})
                continue

            if deactivate:
                self._connection.set_active(False, model, literal_eval(deactivate), search_value_xml_id)
                continue

            if activate:
                self._connection.set_active(True, model, literal_eval(activate), search_value_xml_id)
                continue

            if field_key and not force_id:
                object_ids = self.search(model, [(field_key, '=', values[field_key])],
                                         order='id', context=config_context)
            elif force_id:
                if load:
                    values['id'] = force_id
                else:
                    values, object_ids = self.set_force_id(model, values, config_context, force_id)
            elif action_server_id:
                self.exec_action_server(action_server_id, datas[data], config_context, no_raise)
                continue
            elif function:
                self.exec_function(function, datas[data], config_context, no_raise)
                continue
            else:
                self.logger.error("key=%s", field_key)
                raise
            use_create_write = not force_id or isinstance(force_id, int)
            # prepare many2many list of xmlid
            keys = list(values.keys())
            for key in keys:
                if isinstance(values[key], list) and key.endswith('/id'):
                    if values[key] and isinstance(values[key][0], str) and '.' in values[key][0]:
                        if not force_id or isinstance(force_id, int):
                            field_name = key.replace('/id', '')
                            values[field_name] = [self.get_ref(v) for v in values[key]]
                            values.pop(key)
                        else:
                            field_name = key.replace('/id', '.id')
                            values[field_name] = ','.join([str(self.get_ref(v)) for v in values[key]])
                            values.pop(key)

                elif values[key] and isinstance(values[key], str) and key.endswith('.ids'):
                    field_name = key.replace('.ids', '')
                    many2many_values = self.eval_param_value(values[key], force_safe_eval=True)
                    try:
                        self.execute_odoo(model, 'write', [object_ids, {field_name : many2many_values}])
                    except Exception as err:
                        self.logger.error(err, exc_info=True)
                    values.pop(key)
                elif isinstance(values[key], str) and key.endswith('/id'):
                    value = values[key]
                    if value and '.' in value:
                        if object_ids and not load and use_create_write:
                            field_name = key.replace('/id', '')
                        else:
                            field_name = key.replace('/id', '.id')
                        values[field_name] = int(self.get_ref(value))
                        values.pop(key)
                elif key.endswith('.id') and use_create_write:
                    value = values[key]
                    field_name = key.replace('.id', '')
                    values[field_name] = value
                    values.pop(key)
                elif not isinstance(values[key], str) and key.endswith('/json'):
                    field_name = key.replace('/json', '')
                    values[field_name] = str(json.dumps(values[key]))
                    values.pop(key)

            if load:
                fields, rec_values = self.save_values(model, values, config_context, force_id, object_ids,
                                                      load_batch=load)
                load_fields = list(set(load_fields + fields))
                raw_load_values.append((fields, rec_values[0]))
            else:
                for language in languages:
                    config_context['lang'] = language
                    self.save_values(model, values, config_context, force_id, object_ids)

        if load:
            load_values = [prepare_load_values(load_fields, v[0], v[1]) for v in raw_load_values]
            context = self._context
            self.execute_load(model, load_fields, load_values, context)

    def save_values(self, model, values, config_context, force_id, object_ids, load_batch=False):
        if not force_id or isinstance(force_id, int):
            if object_ids:
                self.execute_odoo(model, 'write', [object_ids, dict(values)], {'context': config_context})
            else:
                self.execute_odoo(model, 'create', [dict(values)], {'context': config_context})
        else:
            load_keys = list(values.keys())
            load_data = [[]]
            for i in load_keys:
                if isinstance(values[i], bool):
                    load_data[0].append(str(values[i]))
                else:
                    load_data[0].append(values[i])
            self.logger.debug('%s: %s %s' % (model, load_keys, load_data))
            if load_batch:
                return load_keys, load_data
            else:
                self.execute_load(model, load_keys, load_data, config_context)

    def execute_load(self, model, load_keys, load_data, config_context):
        res = self.execute_odoo(model, 'load', [load_keys, load_data], {'context': config_context})
        for message in res['messages']:
            self.logger.error("%s : %s" % (message['record'], message['message']))

    def set_force_id(self, model, values, config_context, force_id=''):
        if isinstance(force_id, int):
            values['id'] = force_id
            object_ids = self.search(model, [('id', '=', force_id)], order='id', context=config_context)
        else:
            if '.' not in force_id:
                force_id = "external_config." + force_id
            values['id'] = force_id
            if not self._configurator.xmlid_cache: #load xml_id cache
                self._configurator.get_external_config_xmlid_cache()
            if force_id in self._configurator.xmlid_cache:
                return values, self._configurator.xmlid_cache[force_id]
            module, name = force_id.split('.')
            if self.search('ir.model.data', [('module', '=', module), ('name', '=', name)],
                           order='id', context=config_context):
                object_ids = self._connection.get_ref(force_id)
            else:
                object_ids = []
        return values, object_ids

    def exec_action_server(self, action_server_id, datas, context, no_raise):
        context.update({'active_model': datas.get('model'),
                        'active_id': datas.get('res_id')})
        res = self.execute_odoo('ir.actions.server', 'run', [action_server_id], {'context': context},
                                no_raise=no_raise)

    def eval_param_value(self, param, force_safe_eval=False):
        if isinstance(param, str):
            if param.startswith('get_') or force_safe_eval:
                return self.safe_eval(param, force=force_safe_eval)
            if param[0] in ['[', '{', '(']:
                param_val = literal_eval(param)
                return param_val

        if isinstance(param, OrderedDict):
            return dict(param)
        if isinstance(param, list):
            param_val = []
            for param_el in param:
                param_el_val = self.eval_param_value(param_el)
                param_val.append(param_el_val)
            return param_val

        return param

    def exec_function(self, function, datas, context, no_raise):
        params = datas.get('params')
        if params:
            res_id = datas.get('res_id', False)
            function_params = [res_id] if res_id else []
            for param in params:
                param_val = self.eval_param_value(param)
                function_params.append(param_val)

            extra_params = {'context': context}
            kw = datas.get('kw')
            if kw:
                extra_params.update(dict(kw))

            res = self.execute_odoo(datas.get('model'), function, function_params, extra_params,
                                    no_raise=no_raise)
        else:
            res = self.execute_odoo(datas.get('model'), function, [datas.get('res_id')], {'context': context},
                                    no_raise=no_raise)
